# flake8: noqa
from .archiver import ArchiveManager
from .converters import (
    DraftModel,
    EmulationRunModel,
    FeatureModel,
    FeatureTypeModel,
    PublisherContext,
    ScenarioModel,
    SystemUserModel,
    TagModel,
    TestExecutorContext,
    TestRunModel,
    TestUserSpecification,
)
from .feature import (
    FeatureExtractor,
    FeatureTypeExtractionError,
    FeatureTypeName,
    IFeatureExtractor,
    ScenariosTestFileNotFound,
)
from .file_extractor import BaseFileExtractor
from .git_initializer import GitPullError, GitRepositoryInitializationError, GitRepositoryInitializer
from .language import StepPrefixesModel
from .report_manager import ReportManager, ReportPresenceResolution
from .settings import (
    OverhaveAdminSettings,
    OverhaveDescriptionManagerSettings,
    OverhaveEmulationSettings,
    OverhaveFileSettings,
    OverhaveLanguageSettings,
    OverhaveRedisSettings,
    OverhaveReportManagerSettings,
    OverhaveScenarioCompilerSettings,
    OverhaveStepContextSettings,
    ProcessorSettings,
)



from typing import Optional

from overhave.entities import OverhaveFileSettings, OverhaveLanguageSettings, OverhaveScenarioCompilerSettings
from overhave.factory.context.base_context import BaseFactoryContext
from overhave.publication.gitlab import OverhaveGitlabPublisherSettings, TokenizerClientSettings
from overhave.publication.settings import BaseGitPublisherSettings, PublicationSettings
from overhave.test_execution import OverhaveProjectSettings
from overhave.transport import OverhaveGitlabClientSettings
from overhave.transport.http.base_client.settings import HttpSettingsType


class OverhavePublicationContext(BaseFactoryContext):
    """Overhave publication context, based on application BaseSettings.

    This context defines how Overhave publication will work.
    """

    def __init__(
        self,
        compilation_settings: Optional[OverhaveScenarioCompilerSettings] = None,
        file_settings: Optional[OverhaveFileSettings] = None,
        language_settings: Optional[OverhaveLanguageSettings] = None,
        project_settings: Optional[OverhaveProjectSettings] = None,
        client_settings: Optional[HttpSettingsType] = None,
        publisher_settings: Optional[BaseGitPublisherSettings] = None,
        publication_settings: Optional[PublicationSettings] = None,
        tokenizer_client_settings: Optional[TokenizerClientSettings] = None,
    ) -> None:
        super().__init__(
            compilation_settings=compilation_settings or OverhaveScenarioCompilerSettings(),
            file_settings=file_settings or OverhaveFileSettings(),
            language_settings=language_settings or OverhaveLanguageSettings(),
            project_settings=project_settings or OverhaveProjectSettings(),
        )
        self.client_settings = client_settings or OverhaveGitlabClientSettings()
        self.publisher_settings = publisher_settings or OverhaveGitlabPublisherSettings()
        self.publication_settings = publication_settings or PublicationSettings()
        self.tokenizer_client_settings = tokenizer_client_settings or TokenizerClientSettings()



from overhave import OverhaveAdminApp, overhave_app
from overhave.base_settings import DataBaseSettings, LoggingSettings
from overhave.cli.group import overhave
from overhave.factory import get_admin_factory


def _get_admin_app() -> OverhaveAdminApp:
    DataBaseSettings().setup_db()
    LoggingSettings().setup_logging()
    return overhave_app(get_admin_factory())


@overhave.command(short_help="Run Overhave Admin panel")
def admin(port: int = 8076, debug: bool = False) -> None:
    _get_admin_app().run(host="localhost", port=port, debug=debug)



import abc
from datetime import datetime
from typing import List, Optional, cast

from overhave import db
from overhave.db import DraftStatus
from overhave.entities import DraftModel


class IDraftStorage(abc.ABC):
    """Abstract class for scenario versions storage."""

    @staticmethod
    @abc.abstractmethod
    def get_draft(draft_id: int) -> Optional[DraftModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_drafts_by_feature_id(feature_id: int) -> List[DraftModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_draft(test_run_id: int, published_by: str, status: DraftStatus) -> int:
        pass

    @staticmethod
    @abc.abstractmethod
    def save_response(
        draft_id: int, pr_url: str, published_at: datetime, status: DraftStatus, traceback: Optional[str] = None
    ) -> None:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_previous_feature_draft(feature_id: int) -> DraftModel:
        pass

    @abc.abstractmethod
    def set_draft_status(self, draft_id: int, status: DraftStatus, traceback: Optional[str] = None) -> None:
        pass


class BaseDraftStorageException(Exception):
    """Base exception for :class:`DraftStorage`."""


class UniqueDraftCreationError(BaseDraftStorageException):
    """Exception for draft creation error with `as_unique`."""


class NullableDraftsError(BaseDraftStorageException):
    """Exception for situation with not existing drafts."""


class DraftStorage(IDraftStorage):
    """Class for scenario versions storage."""

    @staticmethod
    def get_draft(draft_id: int) -> Optional[DraftModel]:
        with db.create_session() as session:
            draft: Optional[db.Draft] = session.query(db.Draft).get(draft_id)
            if draft is not None:
                return cast(DraftModel, DraftModel.from_orm(draft))
            return None

    @staticmethod
    def get_drafts_by_feature_id(feature_id: int) -> List[DraftModel]:
        draft_model_list: List[DraftModel] = []
        with db.create_session() as session:
            drafts: List[db.Draft] = session.query(db.Draft).filter(db.Draft.feature_id == feature_id).all()
            for draft in drafts:
                draft_model_list.append(cast(DraftModel, DraftModel.from_orm(draft)))
        return draft_model_list

    @staticmethod
    def save_draft(test_run_id: int, published_by: str, status: DraftStatus) -> int:
        with db.create_session() as session:
            try:
                draft = session.query(db.Draft).as_unique(
                    test_run_id=test_run_id, published_by=published_by, status=status
                )
            except RuntimeError as e:
                raise UniqueDraftCreationError("Could not get unique draft!") from e
            session.add(draft)
            session.flush()
            return cast(int, draft.id)

    @staticmethod
    def save_response(
        draft_id: int, pr_url: str, published_at: datetime, status: DraftStatus, traceback: Optional[str] = None
    ) -> None:
        with db.create_session() as session:
            draft: db.Draft = session.query(db.Draft).get(draft_id)
            draft.pr_url = pr_url
            draft.published_at = published_at
            if traceback is not None:
                draft.traceback = traceback
            draft.status = status
            feature: db.Feature = session.query(db.Feature).get(draft.feature_id)
            feature.released = status.success

    @staticmethod
    def get_previous_feature_draft(feature_id: int) -> DraftModel:
        with db.create_session() as session:
            selection_num = 2
            drafts: List[db.Draft] = (  # noqa: ECE001
                session.query(db.Draft)
                .filter(db.Draft.feature_id == feature_id)
                .order_by(db.Draft.id.asc())
                .limit(selection_num)
                .all()
            )
            if not drafts or len(drafts) != selection_num:
                raise NullableDraftsError(f"Haven't got Drafts amount={selection_num} for feature_id={feature_id}!")
            return cast(DraftModel, DraftModel.from_orm(drafts[0]))

    def set_draft_status(self, draft_id: int, status: DraftStatus, traceback: Optional[str] = None) -> None:
        with db.create_session() as session:
            draft: db.Draft = session.query(db.Draft).get(draft_id)
            draft.status = status
            if not isinstance(traceback, str):
                return
            draft.traceback = traceback



import walrus

from overhave.entities.settings import OverhaveRedisSettings


class RedisTemplate:
    """Base class for :class:`RedisProducer` and :class:`RedisConsumer`."""

    def __init__(self, settings: OverhaveRedisSettings):
        self._settings = settings
        self._database = walrus.Database(
            host=settings.redis_url.host, port=settings.redis_url.port, db=settings.redis_db, retry_on_timeout=True
        )



import abc
from typing import Generic

from overhave.transport import TRedisTask


class ITaskConsumerFactory(Generic[TRedisTask], abc.ABC):
    """Abstract class for factory, which consumes `TRedisTask` tasks."""

    @abc.abstractmethod
    def process_task(self, task: TRedisTask) -> None:
        pass



from typing import List, Mapping, TypeVar

from overhave.base_settings import BaseOverhavePrefix
from overhave.entities import FeatureTypeName
from overhave.publication.objects import PublicationManagerType


class BaseGitPublisherSettings(BaseOverhavePrefix):
    """Base git publisher settings."""

    default_reviewers: List[str] = []
    feature_type_to_reviewers_mapping: Mapping[FeatureTypeName, List[str]] = {}
    default_target_branch_name: str = "master"


class PublicationSettings(BaseOverhavePrefix):
    """Publication settings where you can specify publication manager and its behavior."""

    # Choose gitlab or stash as a publication manager
    publication_manager_type: PublicationManagerType = PublicationManagerType.GITLAB


GitPublisherSettings = TypeVar("GitPublisherSettings", bound=BaseGitPublisherSettings)



import logging
from operator import attrgetter
from typing import Any, Dict, List, Optional, Tuple, cast

from _pytest.fixtures import FixtureDef
from _pytest.main import Session

from overhave.entities import FeatureTypeName, StepPrefixesModel
from overhave.test_execution.objects import BddStepModel

_PYTESTBDD_FIXTURE_MARK = "pytestbdd_"
_PYTESTBDD_FIXTURE_TRACE_MARK = "_trace"

logger = logging.getLogger(__name__)


class BaseStepCollectorException(Exception):
    """Base exception for :class:`StepCollector`."""


class BddStepWithoutDocsError(BaseStepCollectorException):
    """Error for situation when pytest_bdd steps declared without docstring."""


class StepCollector:
    """Class for `pytest-bdd` steps dynamic collection."""

    def __init__(self, step_prefixes: Optional[StepPrefixesModel]) -> None:
        self._step_prefixes = step_prefixes
        self._steps: Dict[FeatureTypeName, List[BddStepModel]] = {}

    @staticmethod
    def _is_bdd_step(fixture: FixtureDef[Any]) -> bool:
        is_bdd_step = (
            isinstance(fixture.argname, str)
            and fixture.argname.startswith(_PYTESTBDD_FIXTURE_MARK)
            and not fixture.argname.endswith(_PYTESTBDD_FIXTURE_TRACE_MARK)
        )
        logger.debug("Fixture: %s - is_bdd_step=%s", fixture.argname, is_bdd_step)
        if is_bdd_step and not isinstance(fixture.func.__doc__, str):
            raise BddStepWithoutDocsError(
                f"Fixture {fixture} does not have description! Please, set it via docstrings."
            )
        return is_bdd_step

    @classmethod
    def _get_pytestbdd_step_fixtures(cls, session: Session) -> Tuple[FixtureDef[Any]]:
        return cast(
            Tuple[FixtureDef[Any]],
            sorted(
                (
                    fx
                    for fx_list in session._fixturemanager._arg2fixturedefs.values()
                    for fx in fx_list
                    if cls._is_bdd_step(fx)
                ),
                key=attrgetter("func.step_type"),
                reverse=True,
            ),
        )

    def _compile_full_step_name(self, fixture_name: str, step_type: str) -> str:
        prefix = step_type.title()
        if self._step_prefixes is not None:
            prefix = self._step_prefixes.dict()[step_type.upper()].strip()
        return f"{prefix} {fixture_name}"

    def _compile_step_models(self, steps: Tuple[FixtureDef[Any]]) -> List[BddStepModel]:
        return [
            BddStepModel(
                type=f.func.step_type,  # type: ignore
                name=self._compile_full_step_name(
                    fixture_name=f.func.parser.name,  # type: ignore
                    step_type=f.func.step_type,  # type: ignore
                ),
                doc=f.func.__doc__,
            )
            for f in steps
        ]

    def collect_steps(self, session: Session, feature_type: FeatureTypeName) -> None:
        logger.debug("Collecting steps for feature_type=%s...", feature_type)
        step_fixtures = self._get_pytestbdd_step_fixtures(session)
        bdd_steps = self._compile_step_models(step_fixtures)
        if bdd_steps:
            logger.debug("Loaded steps dict:\n%s", bdd_steps)
        else:
            logger.warning("Feature type '%s' does not have any pytest_bdd steps!", feature_type)
        self._steps[feature_type] = bdd_steps

    def get_steps(self, feature_type: FeatureTypeName) -> Optional[List[BddStepModel]]:
        return self._steps.get(feature_type)



from pathlib import Path
from typing import List, Optional

import allure
from _pytest.nodes import Item
from pytest_bdd.parser import Scenario

from overhave.pytest_plugin.helpers.basic import get_scenario, is_pytest_bdd_item
from overhave.test_execution import OverhaveProjectSettings


def _get_issue_links(scenario: Scenario, keyword: str) -> Optional[List[str]]:
    keyword_with_colon = keyword.title() + ":"
    with Path(scenario.feature.filename).open() as file:
        for line in file:
            if keyword not in line:
                continue
            links_part = line.split(keyword_with_colon)[-1]
            return [x.strip() for x in links_part.split(",")]
    return None


def set_issue_links(scenario: Scenario, keyword: str) -> None:
    links = _get_issue_links(scenario, keyword)
    if links:
        setattr(scenario.feature, "links", links)


def has_issue_links(item: Item) -> bool:
    return is_pytest_bdd_item(item) and hasattr(get_scenario(item).feature, "links")


def add_issue_links_to_report(project_settings: OverhaveProjectSettings, scenario: Scenario) -> None:
    for link in scenario.feature.links:
        allure.dynamic.link(url=project_settings.get_link_url(link), link_type="issue", name=link)



# flake8: noqa
from .executor import ITestExecutor, TestExecutor
from .objects import BddStepModel, StepTypeName
from .settings import OverhaveProjectSettings, OverhaveTestSettings
from .step_collector import StepCollector
from .test_runner import PytestRunner



# flake8: noqa
from .helpers import (
    DescriptionManager,
    StepContextNotDefinedError,
    StepContextRunner,
    get_description_manager,
    get_scenario,
    get_step_context_runner,
    has_issue_links,
)
from .plugin_resolver import IPluginResolver, PluginResolver
from .proxy_manager import IProxyManager, ProxyManager, get_proxy_manager



import abc
from typing import Optional, cast

from overhave import db
from overhave.entities import TagModel


class IFeatureTagStorage(abc.ABC):
    """Abstract class for feature tags storage."""

    @staticmethod
    @abc.abstractmethod
    def get_or_create_tag(value: str, created_by: str) -> TagModel:
        pass


class FeatureTagStorage(IFeatureTagStorage):
    """Class for feature tags storage."""

    @staticmethod
    def get_or_create_tag(value: str, created_by: str) -> TagModel:
        with db.create_session() as session:
            tag: Optional[db.Tags] = session.query(db.Tags).filter(db.Tags.value == value).one_or_none()
            if tag is None:
                tag = db.Tags(value=value, created_by=created_by)
                session.add(tag)
                session.flush()
            return cast(TagModel, TagModel.from_orm(tag))



from typing import Optional

from pydantic import BaseModel

from overhave.db import TestReportStatus


class ReportPresenceResolution(BaseModel):
    """Model for report presence resolution result."""

    exists: bool
    s3_enabled: bool = False
    report_status: Optional[TestReportStatus]

    @property
    def not_ready(self) -> bool:
        return self.s3_enabled and not self.exists and self.report_status is TestReportStatus.GENERATED



import logging
from typing import Optional

from pydantic import SecretStr

from overhave.authorization.manager.secret_mixin import AdminSecretMixin
from overhave.entities import SystemUserModel

logger = logging.getLogger(__name__)


class DefaultAdminAuthorizationManager(AdminSecretMixin):
    """Class for user authorization.

    Manager authorize users with registered credentials.
    """

    def authorize_user(self, username: str, password: SecretStr) -> Optional[SystemUserModel]:
        return self._system_user_storage.get_user_by_credits(login=username, password=password)



import abc
from typing import Optional, cast

from overhave import db
from overhave.db import create_session
from overhave.entities import FeatureTypeName
from overhave.entities.converters import TestUserModel, TestUserSpecification
from overhave.storage import FeatureTypeStorage


class BaseTestUserStorageException(Exception):
    """Base exception for :class:`FeatureStorage`."""


class TestUserDoesNotExistError(BaseTestUserStorageException):
    """Error for situation when feature not found."""


class ITestUserStorage(abc.ABC):
    """Abstract class for Test User storage."""

    @staticmethod
    @abc.abstractmethod
    def get_test_user_by_id(user_id: int) -> Optional[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def get_test_user_by_name(name: str) -> Optional[TestUserModel]:
        pass

    @staticmethod
    @abc.abstractmethod
    def create_test_user(
        name: str, specification: TestUserSpecification, created_by: str, feature_type: FeatureTypeName
    ) -> TestUserModel:
        pass

    @staticmethod
    @abc.abstractmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        pass


class TestUserStorage(ITestUserStorage):
    """Class for Test User storage."""

    @staticmethod
    def get_test_user_by_id(user_id: int) -> Optional[TestUserModel]:
        with create_session() as session:
            user: Optional[db.TestUser] = session.query(db.TestUser).get(user_id)
            if user is not None:
                return cast(TestUserModel, TestUserModel.from_orm(user))
            return None

    @staticmethod
    def get_test_user_by_name(name: str) -> Optional[TestUserModel]:
        with create_session() as session:
            user: Optional[db.TestUser] = session.query(db.TestUser).filter(db.TestUser.name == name).one_or_none()
            if user is not None:
                return cast(TestUserModel, TestUserModel.from_orm(user))
            return None

    @staticmethod
    def create_test_user(
        name: str, specification: TestUserSpecification, created_by: str, feature_type: FeatureTypeName
    ) -> TestUserModel:
        feature_type = FeatureTypeStorage.get_feature_type_by_name(name=feature_type)
        with db.create_session() as session:
            test_user = db.TestUser(  # type: ignore
                name=name,
                specification=specification,
                feature_type_id=feature_type.id,
                created_by=created_by,
            )
            session.add(test_user)
            session.flush()
            return cast(TestUserModel, TestUserModel.from_orm(test_user))

    @staticmethod
    def update_test_user_specification(user_id: int, specification: TestUserSpecification) -> None:
        with db.create_session() as session:
            test_user = session.query(db.TestUser).get(user_id)
            if test_user is None:
                raise TestUserDoesNotExistError(f"Test user with id {user_id} does not exist!")
            test_user.specification = specification
            session.flush()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['overhave',
 'overhave.admin',
 'overhave.admin.flask',
 'overhave.admin.views',
 'overhave.admin.views.formatters',
 'overhave.admin.views.index',
 'overhave.api',
 'overhave.authorization',
 'overhave.authorization.manager',
 'overhave.authorization.manager.ldap',
 'overhave.cli',
 'overhave.cli.db_cmds',
 'overhave.db',
 'overhave.emulation',
 'overhave.entities',
 'overhave.entities.feature',
 'overhave.entities.language',
 'overhave.entities.report_manager',
 'overhave.extra',
 'overhave.factory',
 'overhave.factory.components',
 'overhave.factory.context',
 'overhave.publication',
 'overhave.publication.gitlab',
 'overhave.publication.gitlab.tokenizer',
 'overhave.publication.stash',
 'overhave.pytest_plugin',
 'overhave.pytest_plugin.helpers',
 'overhave.pytest_plugin.helpers.allure_utils',
 'overhave.scenario',
 'overhave.storage',
 'overhave.synchronization',
 'overhave.test_execution',
 'overhave.transport',
 'overhave.transport.http',
 'overhave.transport.http.base_client',
 'overhave.transport.http.gitlab_client',
 'overhave.transport.http.stash_client',
 'overhave.transport.redis',
 'overhave.transport.s3',
 'overhave.utils']

package_data = \
{'': ['*'], 'overhave.admin': ['files/*', 'files/ace-src/*', 'templates/*']}

install_requires = \
['Flask-WTF>=0.14.2',
 'Flask>=2.0.1,<3.0.0',
 'GitPython>=3.1.15,<4.0.0',
 'SQLAlchemy-Utils>=0.37.0,<0.38.0',
 'WTForms>=2.2',
 'alembic>=1.4.3,<2.0.0',
 'boto3-type-annotations>=0.3.1,<0.4.0',
 'boto3>=1.17.16,<2.0.0',
 'docker',
 'fastapi>=0.75.0,<0.76.0',
 'flask-admin>=1.5,<2.0',
 'flask-login>=0.4.1,<0.5.0',
 'gunicorn>=20.0.4,<21.0.0',
 'httptools>=0.1.1,<0.2.0',
 'ldap3>=2.6,<3.0',
 'psycopg2-binary>=2.8,<3.0',
 'pydantic-sqlalchemy>=0.0.9,<0.0.10',
 'pydantic>=1.7',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-gitlab>=2.9.0,<3.0.0',
 'python-ldap>=3.2,<4.0',
 'pytz',
 'redis>=3.4.1,<4.0.0',
 'requests>=2.0.0',
 'sqlalchemy-utc>=0.14.0,<0.15.0',
 'sqlalchemy>=1.3.23',
 'tenacity',
 'typer>=0.4.0,<0.5.0',
 'walrus>=0.8.0,<0.9.0',
 'wsgi_intercept>=1.8,<2.0',
 'yarl>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['overhave = overhave.cli:overhave',
                     'overhave-demo = demo:overhave_demo'],
 'pytest11': ['overhave = overhave.pytest_plugin.plugin']}

setup_kwargs = {
    'name': 'overhave',
    'version': '2.8.1',
    'description': 'Overhave - web-framework for BDD',
    'long_description': '========\nOverhave\n========\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/label_img.png\n  :width: 700\n  :align: center\n  :alt: Overhave framework\n\n  `Overhave`_ is the web-framework for BDD: scalable, configurable, easy to use, based on\n  `Flask Admin`_ and `Pydantic`_.\n\n  .. image:: https://github.com/TinkoffCreditSystems/overhave/workflows/CI/badge.svg\n    :target: https://github.com/TinkoffCreditSystems/overhave/actions?query=workflow%3ACI\n    :alt: CI\n\n  .. image:: https://img.shields.io/pypi/pyversions/overhave.svg\n    :target: https://pypi.org/project/overhave\n    :alt: Python versions\n\n  .. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/TinkoffCreditSystems/overhave\n    :alt: Code style\n\n  .. image:: https://img.shields.io/pypi/v/overhave?color=%2334D058&label=pypi%20package\n    :target: https://pypi.org/project/overhave\n    :alt: Package version\n    \n  .. image:: https://img.shields.io/pypi/dm/overhave.svg\n    :target: https://pypi.org/project/overhave\n    :alt: Downloads per month\n\n--------\nFeatures\n--------\n\n* Ready web-interface for easy BDD features management with `Ace`_ editor\n* Traditional Gherkin format for scenarios provided by `pytest-bdd`_\n* Execution and reporting of BDD features based on `PyTest`_  and `Allure`_\n* Auto-collection of `pytest-bdd`_ steps and display on the web-interface\n* Simple business-alike scenarios structure, easy horizontal scaling\n* Built-in wrappers for `pytest-bdd`_ hooks to supplement `Allure`_ report\n* Ability to create and use several BDD keywords dictionary with different languages\n* Versioning and deployment of scenario drafts to `Bitbucket`_ or `GitLab`_\n* Synchronization between `git`_ repository and database with features\n* Built-in configurable management of users and groups permissions\n* Configurable strategy for user authorization, LDAP also provided\n* Database schema based on `SQLAlchemy`_ models and works with PostgreSQL\n* Still configurable as `Flask Admin`_, supports plug-ins and extensions\n* Distributed `producer-consumer` architecture based on Redis streams\n  through `Walrus`_\n* Web-browser emulation ability with custom toolkit (`GoTTY`_, for example)\n* Simple command-line interface, provided with `Typer`_\n* Integrated interaction for files storage with s3-cloud based on `boto3`_\n\n------------\nInstallation\n------------\n\nYou can install **Overhave** via pip from PyPI:\n\n.. code-block:: shell\n\n    pip install overhave\n\n--------\nOverview\n--------\n\nWeb-interface\n-------------\n\nThe web-interface is a basic tool for BDD features management. It consists of:\n\n* `Info` - index page with optional information about your tool or project;\n* `Scenarios` - section for features management, contains subsections\n    `Features`, `Test runs` and `Versions`:\n\n    * `Features`\n        gives an interface for features records management and provides info\n        about id, name author, time, editor and publishing status; it is possible\n        to search, edit or delete items through Script panel.\n\n        .. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/label_img.png\n          :width: 500\n          :align: center\n          :alt: Features list\n\n    * `Test runs`\n        gives an interface for test runs management and provides info about.\n\n        .. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/test_runs_img.png\n          :width: 500\n          :align: center\n          :alt: Test runs list\n\n    * Versions\n        contains feature versions in corresponding to test runs; versions contains PR-links to\n        the remote Git repository (only Stash is supported now).\n\n        .. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/versions_img.png\n          :width: 500\n          :align: center\n          :alt: Feature published versions list\n\n    * Tags\n        contains tags values, which are used for feature\'s tagging.\n\n        .. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/tags_img.png\n          :width: 500\n          :align: center\n          :alt: Feature published versions list\n\n* `Access` - section for access management, contains `Users` and\n    `Groups` subsections;\n* `Emulation` - experimental section for alternative tools implementation\n    (in development).\n\n**Overhave** features could be created and/or edited through special\n*script panel* in feature edit mode. Feature should have type registered by the\napplication, unique name, specified tasks list with the traditional format\n```PRJ-NUMBER``` and scenario text.\n\n**Script panel** has `pytest-bdd`_ steps table on the right side of interface.\nThese steps should be defined in appropriate fixture modules and registered\nat the application on start-up to be displayed.\n\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/panel_img.png\n  :width: 600\n  :align: center\n  :alt: Script panel\n\n  Example of **Overhave** script panel in feature edit mode\n\nAllure report\n-------------\n\n**Overhave** generates `Allure`_ report after tests execution in web-interface.\nIf you execute tests manually through `PyTest`_, these results are could be\nconverted into the `Allure`_ report also with the `Allure CLI`_ tool.\nThis report contains scenarios descriptions as they are described in features.\n\n.. figure:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/report_img.png\n  :width: 600\n  :align: center\n  :alt: Allure test-case report\n\n  Example of generated `Allure`_ report after execution of **Overhave**\'s feature\n\nDemo-mode (Quickstart)\n----------------------\n\n**Overhave** has special demo-mode (in development), which could be possibly\nused for framework demonstration and manual debugging / testing. The framework\nprovides a CLI entrypoints for easy server run in debug mode:\n\n.. code-block:: shell\n\n    make up  # start PostgreSQL database and Redis\n    overhave db create-all  # create Overhave database schema\n    overhave-demo admin  # start Overhave admin on port 8076 in debug mode\n    overhave-demo consumer -s test  # start Overhave test execution consumer\n\n**Note**: you could run admin in special mode, which does not require\nconsumers. This mode uses *threadpool* for running testing and publication\ntasks asynchronously:\n\n.. code-block:: shell\n\n    overhave-demo admin --threadpool --language=ru\n\nBut this *threadpool* mode is unscalable in *kubernetes* paradigm. So,\nit\'s highly recommended to use corresponding consumers exactly.\n\nCommand-line interface\n----------------------\n\n**Overhave** has a CLI that provides a simple way to start service web-interface,\nrun consumer and execute basic database operations. Examples are below:\n\n.. code-block:: shell\n\n    overhave db create-all\n    overhave admin --port 8080\n    overhave consumer -s publication\n\n**Note**: service start-up takes a set of settings, so you can set them through\nvirtual environment with prefix ```OVERHAVE_```, for example ```OVERHAVE_DB_URL```.\nIf you want to configure settings in more explicit way through context injection,\nplease see next part of docs.\n\nContext injection\n-----------------\n\nContext setting\n^^^^^^^^^^^^^^^\n\nService could be configured via application context injection with prepared\ninstance of `OverhaveContext` object. This context could be set using\n```set_context``` function of initialized ```ProxyFactory``` instance.\n\nFor example, ```my_custom_context``` prepared. So, application start-up could\nbe realised with follow code:\n\n.. code-block:: python\n\n    from overhave import overhave_app, overhave_admin_factory\n\n    factory = overhave_admin_factory()\n    factory.set_context(my_custom_context)\n    overhave_app(factory).run(host=\'localhost\', port=8080, debug=True)\n\n**Note**:\n\n* ```overhave_app``` is the prepared `Flask` application with already enabled\n    Flask Admin and Login Manager plug-ins;\n* ```overhave_factory``` is a function for LRU cached instance of the **Overhave**\n    factory ```ProxyFactory```; the instance has an access to application components,\n    directly used in ```overhave_app```.\n* ```my_custom_context``` is an example of context configuration, see an\n    example code in `context_example.rst`_.\n\nEnabling of injection\n^^^^^^^^^^^^^^^^^^^^^\n\n**Overhave** has it\'s own built-in `PyTest`_ plugin, which is used to enable\nand configure injection of prepared context into application core instance.\nThe plugin provides one option:\n\n* `--enable-injection` - flag to enable context injection.\n\nThe `PyTest` usage should be similar to:\n\n.. code-block:: bash\n\n    pytest --enable-injection\n\n\nConsumers\n---------\n\n**Overhave** has `producer-consumer` architecture, based on Redis streams,\nand supported 3 consumer\'s types:\n\n* **TEST** - consumer for test execution with it\'s own factory\n    ```overhave_test_execution_factory```;\n\n* **PUBLICATION** - consumer for features publication with it\'s own factory\n    ```overhave_publication_factory```;\n\n* **EMULATION** - consumer for specific emulation with it\'s own factory\n    ```overhave_emulation_factory```.\n\n**Note**: the ```overhave_test_execution_factory``` has ability for context injection\nand could be enriched with the custom context as the ```overhave_admin_factory```.\n\n\nProject structure\n-----------------\n\n**Overhave** supports it\'s own special project structure:\n\n.. image:: https://raw.githubusercontent.com/TinkoffCreditSystems/overhave/master/docs/includes/images/project_structure.png\n  :width: 300\n  :alt: **Overhave** project structure\n\nThe right approach is to create a **root directory** (like "demo" inside the current\nrepository) that contains **features**, **fixtures** and **steps** directories.\n\nThe **Features** directory contains different feature types as\nseparate directories, each of them corresponds to predefined `pytest-bdd`_\nset of steps.\n\nThe **Fixtures** directory contains typical `PyTest`_ modules splitted by different\nfeature types. These modules are used for `pytest-bdd`_ isolated test runs. It is\nnecessary because of special mechanism of `pytest-bdd`_ steps collection.\n\nThe **Steps** directory contains `pytest-bdd`_ steps packages splitted by differrent\nfeature types also. Each steps subdirectory has it\'s own declared steps in according\nto supported feature type.\n\nSo, it is possible to create your own horizontal structure of\ndifferent product directions with unique steps and `PyTest`_ fixtures.\n\n**Note**: this structure is used in **Overhave** application. The formed data\ngives a possibility to specify registered feature type in the web-interface\n*script panel*. Also, this structure defines which steps will be displayed in\nthe right side of *script panel*.\n\nFeature format\n--------------\n\n**Overhave** has it\'s own special feature\'s text format, which inherits\nGherkin from `pytest-bdd`_ with small updates:\n\n* required tag that is related to existing feature type directory, where\n    current feature is located;\n* info about feature - who is creator, last editor and publisher;\n* task tracker\'s tickets with traditional format ```PRJ-NUMBER```.\n\nAn example of filled feature content is located in\n`feature_example.rst`_.\n\nLanguage\n--------\n\nThe web-interface language is ENG by default and could not be switched\n(if it\'s necessary - please, create a ```feature request``` or contribute\nyourself).\n\nThe feature text as well as `pytest-bdd`_ BDD keywords are configurable\nwith **Overhave** extra models, for example RUS keywords are already defined\nin framework and available for usage:\n\n.. code-block:: python\n\n    from overhave.extra import RUSSIAN_PREFIXES\n\n    language_settings = OverhaveLanguageSettings(step_prefixes=RUSSIAN_PREFIXES)\n\n**Note**: you could create your own prefix-value mapping for your language:\n\n.. code-block:: python\n\n    from overhave import StepPrefixesModel\n\n    GERMAN_PREFIXES = StepPrefixesModel(\n        FEATURE="Merkmal:",\n        SCENARIO_OUTLINE="Szenarioübersicht:",\n        SCENARIO="Szenario:",\n        BACKGROUND="Hintergrund:",\n        EXAMPLES="Beispiele:",\n        EXAMPLES_VERTICAL="Beispiele: Vertikal",\n        GIVEN="Gegeben ",\n        WHEN="Wann ",\n        THEN="Dann ",\n        AND="Und ",\n        BUT="Aber ",\n    )\n\n\nGit integration\n---------------\n\n**Overhave** gives an ability to sent your new features or changes to\nremote git repository, which is hosted by `Bitbucket`_ or `GitLab`_.\nIntegration with bitbucket is native, while integration with GitLab\nuses `python-gitlab`_ library.\n\nYou are able to set necessary settings for your project:\n\n.. code-block:: python\n\n    publisher_settings = OverhaveGitlabPublisherSettings(\n        repository_id=\'123\',\n        default_target_branch_name=\'master\',\n    )\n    client_settings=OverhaveGitlabClientSettings(\n        url="https://gitlab.mycompany.com",\n        auth_token=os.environ.get("MY_GITLAB_AUTH_TOKEN"),\n    )\n\nThe pull-request (for Bitbucket) or merge-request (for GitLab)\ncreated when you click the button `Create pull request` on\ntest run result\'s page. This button is available only for `success`\ntest run\'s result.\n\n**Note**: one of the most popular cases of GitLab API\nauthentication is the OAUTH2 schema with service account.\nIn according to this schema, you should have OAUTH2 token,\nwhich is might have a short life-time and could not be\nspecified through environment. For this situation, **Overhave**\nhas special `TokenizerClient` with it\'s own\n`TokenizerClientSettings` - this simple client could take\nthe token from a remote custom GitLab tokenizer service.\n\n\nGit-to-DataBase synchronization\n-------------------------------\n\n**Overhave** gives an ability to synchronize your current `git`_\nrepository\'s state with database. It means that your features,\nwhich are located on the database, could be updated - and the source\nof updates is your repository.\n\n**For example**: you had to do bulk data replacement in `git`_\nrepository, and now you want to deliver changes to remote database.\nThis not so easy matter could be solved with **Overhave** special\ntooling:\n\nYou are able to set necessary settings for your project:\n\n.. code-block:: bash\n\n    overhave synchronize  # only update existing features\n    overhave synchronize --create-db-features  # update + create new features\n\nYou are able to test this tool with **Overhave** demo mode.\nBy default, 3 features are created in demo database. Just try\nto change them or create new features and run synchronization\ncommand - you will get the result.\n\n.. code-block:: bash\n\n    overhave-demo synchronize  # or with \'--create-db-features\'\n\n\nCustom index\n------------\n\n**Overhave** gives an ability to set custom index.html file for rendering. Path\nto file could be set through environment as well as set with context:\n\n.. code-block:: python\n\n    admin_settings = OverhaveAdminSettings(\n        index_template_path="/path/to/index.html"\n    )\n\n\nAuthorization strategy\n----------------------\n\n**Overhave** provides several authorization strategies, declared by\n```AuthorizationStrategy``` enum:\n\n* `Simple` - strategy without real authorization.\n    Each user could use preferred name. This name will be used for user\n    authority. Each user is unique. Password not required.\n\n* `Default` - strategy with real authorization.\n    Each user could use only registered credentials.\n\n* `LDAP` - strategy with authorization using remote LDAP server.\n    Each user should use his LDAP credentials. LDAP\n    server returns user groups. If user in default \'admin\' group or his groups\n    list contains admin group - user will be authorized. If user already placed\n    in database - user will be authorized too. No one password stores.\n\nAppropriate strategy and additional data should be placed into\n```OverhaveAuthorizationSettings```, for example LDAP strategy could be\nconfigured like this:\n\n.. code-block:: python\n\n    auth_settings=OverhaveAuthorizationSettings(\n        auth_strategy=AuthorizationStrategy.LDAP, admin_group="admin"\n    )\n\nS3 cloud\n--------\n\n**Overhave** implements functionality for *s3* cloud interactions, such as\nbucket creation and deletion, files uploading, downloading and deletion.\nThe framework provides an ability to store reports and other files in\nthe remote s3 cloud storage. You could enrich your environment with following\nsettings:\n\n.. code-block:: shell\n\n    OVERHAVE_S3_ENABLED=true\n    OVERHAVE_S3_URL=https://s3.example.com\n    OVERHAVE_S3_ACCESS_KEY=<MY_ACCESS_KEY>\n    OVERHAVE_S3_SECRET_KEY=<MY_SECRET_KEY>\n\nOptionally, you could change default settings also:\n\n.. code-block:: shell\n\n    OVERHAVE_S3_VERIFY=false\n    OVERHAVE_S3_AUTOCREATE_BUCKETS=true\n\nThe framework with enabled ```OVERHAVE_S3_AUTOCREATE_BUCKETS``` flag will create\napplication buckets in remote storage if buckets don\'t exist.\n\n------------\nContributing\n------------\n\nContributions are very welcome.\n\nPreparation\n-----------\n\nProject installation is very easy\nand takes just few prepared commands (`make pre-init` works only for Ubuntu;\nso you can install same packages for your OS manually):\n\n.. code-block:: shell\n\n    make pre-init\n    make init\n\nPackages management is provided by `Poetry`_.\n\nCheck\n-----\n\nTests can be run with `Tox`_. `Docker-compose`_ is used for other services\npreparation and serving, such as database. Simple tests and linters execution:\n\n.. code-block:: shell\n\n    make up\n    make test\n    make lint\n\nPlease, see `make` file and discover useful shortcuts. You could run tests\nin docker container also:\n\n.. code-block:: shell\n\n    make test-docker\n\nDocumentation build\n-------------------\n\nProject documentation could be built via `Sphinx`_ and simple `make` command:\n\n.. code-block:: shell\n\n    make build-docs\n\nBy default, the documentation will be built using `html` builder into `_build`\ndirectory.\n\n-------\nLicense\n-------\n\nDistributed under the terms of the `GNU GPLv2`_ license.\n\n------\nIssues\n------\n\nIf you encounter any problems, please report them here in section `Issues`\nwith a detailed description.\n\n.. _`Overhave`: https://github.com/TinkoffCreditSystems/overhave\n.. _`Pydantic`: https://github.com/samuelcolvin/pydantic\n.. _`Flask Admin`: https://github.com/flask-admin/flask-admin\n.. _`Ace`: https://github.com/ajaxorg/ace\n.. _`PyTest`: https://github.com/pytest-dev/pytest\n.. _`pytest-bdd`: https://github.com/pytest-dev/pytest-bdd\n.. _`Allure`: https://github.com/allure-framework/allure-python\n.. _`Allure CLI`: https://docs.qameta.io/allure/#_get_started\n.. _`Bitbucket`: https://www.atlassian.com/git\n.. _`GitLab`: https://about.gitlab.com\n.. _`python-gitlab`: https://python-gitlab.readthedocs.io\n.. _`SQLAlchemy`: https://github.com/sqlalchemy/sqlalchemy\n.. _`Walrus`: https://github.com/coleifer/walrus\n.. _`GoTTY`: https://github.com/yudai/gotty\n.. _`GNU GPLv2`: http://www.apache.org/licenses/LICENSE-2.0\n.. _`Tox`: https://github.com/tox-dev/tox\n.. _`Poetry`: https://github.com/python-poetry/poetry\n.. _`Docker-compose`: https://docs.docker.com/compose\n.. _`Typer`: https://github.com/tiangolo/typer\n.. _`Sphinx`: https://github.com/sphinx-doc/sphinx\n.. _`boto3`: https://github.com/boto/boto3\n.. _`git`: https://git-scm.com/\n.. _`context_example.rst`: https://github.com/TinkoffCreditSystems/overhave/blob/master/docs/includes/context_example.rst\n.. _`feature_example.rst`: https://github.com/TinkoffCreditSystems/overhave/blob/master/docs/includes/features_structure_example/feature_type_1/full_feature_example_en.feature\n',
    'author': 'Vladislav Mukhamatnurov',
    'author_email': 'livestreamepidemz@yandex.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging
from http import HTTPStatus

from gitlab import GitlabCreateError, GitlabHttpError
from gitlab.v4.objects.merge_requests import ProjectMergeRequest
from requests import HTTPError

from overhave.db import DraftStatus
from overhave.entities import GitRepositoryInitializer, OverhaveFileSettings, PublisherContext
from overhave.publication.git_publisher import GitVersionPublisher
from overhave.publication.gitlab.settings import OverhaveGitlabPublisherSettings
from overhave.publication.gitlab.tokenizer.client import TokenizerClient
from overhave.scenario import FileManager
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.test_execution import OverhaveProjectSettings
from overhave.transport.http.gitlab_client import GitlabHttpClient, GitlabMrRequest

logger = logging.getLogger(__name__)


class BaseGitlabVersionPublisherException(Exception):
    """Base exception for :class:`GitlabVersionPublisher`."""


class InvalidWebUrlException(BaseGitlabVersionPublisherException):
    """Exception for case when web url is None."""


class GitlabVersionPublisher(GitVersionPublisher[OverhaveGitlabPublisherSettings]):
    """Class for feature version's merge requests management relative to Gitlab API."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        git_initializer: GitRepositoryInitializer,
        gitlab_publisher_settings: OverhaveGitlabPublisherSettings,
        gitlab_client: GitlabHttpClient,
        tokenizer_client: TokenizerClient,
    ):
        super().__init__(
            file_settings=file_settings,
            project_settings=project_settings,
            feature_storage=feature_storage,
            scenario_storage=scenario_storage,
            test_run_storage=test_run_storage,
            draft_storage=draft_storage,
            file_manager=file_manager,
            git_initializer=git_initializer,
        )
        self._gitlab_publisher_settings = gitlab_publisher_settings
        self._gitlab_client = gitlab_client
        self._tokenizer_client = tokenizer_client

    def publish_version(self, draft_id: int) -> None:
        logger.info("Start processing draft_id=%s...", draft_id)
        self._draft_storage.set_draft_status(draft_id, DraftStatus.CREATING)
        context = self._push_version(draft_id)
        if not isinstance(context, PublisherContext):
            return
        merge_request = GitlabMrRequest(
            project_id=self._gitlab_publisher_settings.repository_id,
            title=context.feature.name,
            source_branch=context.target_branch,
            target_branch=self._gitlab_publisher_settings.target_branch,
            description=self._compile_publication_description(context),
            reviewer_ids=self._gitlab_publisher_settings.get_reviewers(feature_type=context.feature.feature_type.name),
        )
        logger.info("Prepared merge-request: %s", merge_request.json(by_alias=True))
        try:
            token = (
                self._gitlab_client._settings.auth_token or self._tokenizer_client.get_token(draft_id=draft_id).token
            )
            response = self._gitlab_client.send_merge_request(
                merge_request=merge_request, token=token, repository_id=self._gitlab_publisher_settings.repository_id
            )
            if isinstance(response, ProjectMergeRequest):
                if response.web_url is None:
                    raise InvalidWebUrlException("Please verify your gitlab url environment! It is invalid!")
                self._draft_storage.save_response(
                    draft_id=draft_id,
                    pr_url=response.web_url,
                    published_at=response.created_at,
                    status=DraftStatus.CREATED,
                )
                return
        except (GitlabCreateError, GitlabHttpError) as e:
            if e.response_code == HTTPStatus.CONFLICT:
                context.draft.traceback = str(e)
                logger.exception("Gotten conflict. Try to return last merge-request for Draft with id=%s...", draft_id)
                self._save_as_duplicate(context)
                return
            self._draft_storage.set_draft_status(draft_id, DraftStatus.INTERNAL_ERROR, traceback=str(e))
        except HTTPError as e:
            self._draft_storage.set_draft_status(draft_id, DraftStatus.INTERNAL_ERROR, traceback=str(e))
            logger.exception("Got HTTP error while trying to sent merge-request!")



import logging
from http import HTTPStatus
from typing import Optional

import fastapi

from overhave.api.deps import get_test_user_storage
from overhave.entities.converters import TestUserModel
from overhave.storage import ITestUserStorage

logger = logging.getLogger(__name__)


def _test_user_id_handler(user_id: int, test_user_storage: ITestUserStorage) -> TestUserModel:
    logger.info("Getting %s with user_id=%s...", TestUserModel.__name__, user_id)
    test_user = test_user_storage.get_test_user_by_id(user_id)
    if test_user is None:
        raise fastapi.HTTPException(status_code=HTTPStatus.NOT_FOUND, detail=f"User with id={user_id} does not exist")
    return test_user


def _test_user_name_handler(user_name: str, test_user_storage: ITestUserStorage) -> TestUserModel:
    logger.info("Getting %s with user_name='%s'...", TestUserModel.__name__, user_name)
    test_user = test_user_storage.get_test_user_by_name(user_name)
    if test_user is None:
        raise fastapi.HTTPException(
            status_code=HTTPStatus.NOT_FOUND, detail=f"User with name='{user_name}' does not exist"
        )
    return test_user


def test_user_handler(
    user_id: Optional[int] = None,
    user_name: Optional[str] = None,
    test_user_storage: ITestUserStorage = fastapi.Depends(get_test_user_storage),
) -> TestUserModel:
    if user_id is not None:
        return _test_user_id_handler(user_id=user_id, test_user_storage=test_user_storage)
    if user_name is not None:
        return _test_user_name_handler(user_name=user_name, test_user_storage=test_user_storage)
    raise fastapi.HTTPException(
        status_code=HTTPStatus.BAD_REQUEST, detail="'user_id' or 'user_name' query parameter should be set"
    )



import abc
import logging
from typing import Generic, Optional, cast

import git

from overhave.entities import GitPullError, GitRepositoryInitializer, OverhaveFileSettings, PublisherContext
from overhave.publication.base_publisher import BaseVersionPublisher, BaseVersionPublisherException
from overhave.publication.settings import GitPublisherSettings
from overhave.scenario import FileManager
from overhave.storage import IDraftStorage, IFeatureStorage, IScenarioStorage, ITestRunStorage
from overhave.test_execution import OverhaveProjectSettings

logger = logging.getLogger(__name__)


class BaseGitVersionPublisherError(BaseVersionPublisherException):
    """Base exception for git version publisher."""


class CurrentBranchEqualToDefaultError(BaseGitVersionPublisherError):
    """Exception for situation when current branch is equal to default branch."""


class CurrentBranchNotEqualToTargetError(BaseGitVersionPublisherError):
    """Exception for situation when current branch is not equal to target branch."""


class CommitNotCreatedError(BaseGitVersionPublisherError):
    """Exception for situation with not created commit."""


class GitVersionPublisher(Generic[GitPublisherSettings], BaseVersionPublisher, abc.ABC):
    """Class for feature version's management, based on git."""

    def __init__(
        self,
        file_settings: OverhaveFileSettings,
        project_settings: OverhaveProjectSettings,
        feature_storage: IFeatureStorage,
        scenario_storage: IScenarioStorage,
        test_run_storage: ITestRunStorage,
        draft_storage: IDraftStorage,
        file_manager: FileManager,
        git_initializer: GitRepositoryInitializer,
    ) -> None:
        super().__init__(
            file_settings=file_settings,
            project_settings=project_settings,
            feature_storage=feature_storage,
            scenario_storage=scenario_storage,
            test_run_storage=test_run_storage,
            draft_storage=draft_storage,
            file_manager=file_manager,
        )
        self._git_initializer = git_initializer

    def _create_head(self, name: str) -> git.SymbolicReference:
        repo = self._git_initializer.repository
        if name in repo.heads:
            repo.delete_head(repo.heads[name], force=True)
            logger.warning("Trashed existed branch with name '%s'", name)

        branch = repo.create_head(name)
        logger.info("Created head for branch: %s", branch)
        assert repo.active_branch != branch, f"Active branch '{repo.active_branch.name}' is not '{branch.name}'!"
        return branch

    def _create_commit_and_push(self, context: PublisherContext, target_branch: git.Head) -> None:
        repo = self._git_initializer.repository
        original_branch = repo.active_branch
        changes_pushed = False
        try:
            target_branch.checkout()
            logger.info("Active branch after checkout: %s", repo.active_branch)

            feature_file = self._file_manager.produce_feature_file(context=context)
            logger.info("Untracked files after feature creation: %s", repo.untracked_files)
            repo.index.add([feature_file.as_posix()])
            logger.info("Untracked files after index update: %s", repo.untracked_files)

            repo.index.commit(
                (
                    f"Added feature {context.feature.name} "
                    f"(id {context.feature.id}) "
                    f"by {context.test_run.executed_by}"
                ),
                skip_hooks=True,
            )
            logger.info("Commit successfully created")
            if repo.active_branch == original_branch:
                raise CurrentBranchEqualToDefaultError(
                    f"Current branch could not be equal to default: '{original_branch}'!"
                )
            if repo.active_branch != target_branch:
                raise CurrentBranchNotEqualToTargetError(
                    f"Current branch should be equal to target: '{target_branch}'!"
                )

            logger.info("Remote origin url: %s", repo.remotes.origin.url)
            repo.git.push("origin", target_branch, force=True)
            logger.info("Changes successfully pushed to remote repository")
            changes_pushed = True
        except (git.GitCommandError, BaseGitVersionPublisherError):
            logger.exception("Error while trying to commit feature file!")

        if original_branch != repo.active_branch:
            logger.info("Current branch is '%s'", repo.active_branch)
            original_branch.checkout()
            logger.info("Branch returned to '%s'", repo.active_branch)

        if not changes_pushed:
            raise CommitNotCreatedError("Commit has not been created!")

    def _push_version(self, draft_id: int) -> Optional[PublisherContext]:
        try:
            self._git_initializer.pull()
            ctx = self._compile_context(draft_id)
            git_branch = cast(git.Head, self._create_head(name=ctx.target_branch))
            self._create_commit_and_push(context=ctx, target_branch=git_branch)
            return ctx
        except GitPullError:
            logger.exception("Error while trying to pull remote repository!")
        except (git.GitCommandError, BaseGitVersionPublisherError):
            logger.exception("Error while trying to push scenario version!")
        return None



import os
import pathlib

from overhave.base_settings import OVERHAVE_ENV_PREFIX

worker_class = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_WORKER_CLASS", "uvicorn.workers.UvicornWorker")
workers = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_WORKERS", 1)
threads = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_THREADS", 1)
worker_connections = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_CONNECTIONS", 1000)
timeout = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_TIMEOUT", 10)
max_requests = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_REQUESTS", 10000)
backlog = int(os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_BACKLOG", 2048))
max_requests_jitter = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_REQUESTS_JITTER", 100)
reuse_port = True

port = os.environ.get(OVERHAVE_ENV_PREFIX + "GUNICORN_PORT", 8000)
bind = f"0.0.0.0:{port}"

if pathlib.Path("/dev/shm").exists():
    worker_tmp_dir = "/dev/shm"

errorlog = "-"
accesslog = "-"



from typing import Any, cast

from flask import redirect, request, url_for
from flask_admin.contrib.sqla import ModelView
from flask_login import current_user

from overhave.admin.views.formatters.formatters import (
    datetime_formatter,
    draft_feature_formatter,
    draft_prurl_formatter,
    draft_testrun_formatter,
    feature_link_formatter,
    file_path_formatter,
    json_formatter,
    result_report_formatter,
    task_formatter,
)


class ModelViewConfigured(ModelView):
    """Base model view."""

    column_default_sort = ("id", True)
    column_display_pk = True
    can_view_details = True
    page_size = 100
    simple_list_pager = True

    column_exclude_list = ("meta",)
    column_labels = {
        "user": "Author",
        "name": "Name",
        "feature_type": "Type",
        "task": "Tasks",
        "last_edited_by": "Editor",
        "released": "Published",
        "created_at": "Creation time",
        "scenario": "Scenario",
        "start": "Start time",
        "end": "Finish time",
        "executed_by": "Initiator",
        "status": "Status",
        "report": "Report",
        "traceback": "Error traceback",
        "feature_id": "Feature",
        "test_run_id": "Test run ID",
        "pr_url": "Pull-request URL",
        "initiated_by": "Initiator",
        "published_by": "Publisher",
    }
    column_formatters = {
        "created_at": datetime_formatter,
        "start": datetime_formatter,
        "end": datetime_formatter,
        "task": task_formatter,
        "status": result_report_formatter,
        "specification": json_formatter,
        "name": feature_link_formatter,
        "feature_id": draft_feature_formatter,
        "test_run_id": draft_testrun_formatter,
        "pr_url": draft_prurl_formatter,
        "file_path": file_path_formatter,
    }
    column_descriptions = dict(
        name="Feature header for business scenarios",
        task="Tracker tasks converted to web links when viewed",
        last_edited_by="Last editor of scenarios set",
        author="Author of scenarios set",
        released="Feature's last version publishing status",
        executed_by="Initiator of scenarios set test run",
        status="Test run result",
    )

    def is_accessible(self) -> Any:
        return current_user.is_authenticated

    def inaccessible_callback(self, name: str, **kwargs: Any) -> Any:
        # redirect to login page if user doesn't have access
        return redirect(url_for("admin.login", next=request.url))

    @property
    def current_user(self) -> str:
        return cast(str, current_user.login)


class ModelViewProtected(ModelViewConfigured):
    """Protected view."""

    can_delete = False
    can_edit = False
    can_create = False



import logging
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

from overhave.entities import IFeatureExtractor, OverhaveFileSettings, TestExecutorContext
from overhave.scenario.compiler import ScenarioCompiler
from overhave.test_execution.settings import OverhaveProjectSettings

logger = logging.getLogger(__name__)


class FileSavingMixin:
    """Mixin for files saving."""

    @staticmethod
    def _write_data(file: tempfile._TemporaryFileWrapper, data: str, entity_name: str) -> None:  # type: ignore
        file.write(data)
        file.flush()
        logger.debug("Generated %s file: '%s'", entity_name, file.name)
        assert Path(file.name).exists(), f"{entity_name.upper()} file not created!"


class FileManager(FileSavingMixin):
    """Class for feature files management."""

    def __init__(
        self,
        project_settings: OverhaveProjectSettings,
        file_settings: OverhaveFileSettings,
        feature_extractor: IFeatureExtractor,
        scenario_compiler: ScenarioCompiler,
    ):
        self._project_settings = project_settings
        self._file_settings = file_settings
        self._feature_extractor = feature_extractor
        self._scenario_compiler = scenario_compiler

    @contextmanager
    def tmp_feature_file(
        self, context: TestExecutorContext
    ) -> Iterator[tempfile._TemporaryFileWrapper]:  # type: ignore
        file_name = Path(context.feature.file_path).name
        logger.debug("Feature file name: '%s'", file_name)
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_features_dir,
            prefix=f"{file_name}_id{context.feature.id}",
            suffix=self._file_settings.feature_suffix,
            mode="w",
        ) as file:
            data = self._scenario_compiler.compile(context=context)
            logger.debug("Scenario file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="feature")
            yield file

    @contextmanager
    def tmp_fixture_file(
        self, context: TestExecutorContext, feature_file: tempfile._TemporaryFileWrapper  # type: ignore
    ) -> Iterator[tempfile._TemporaryFileWrapper]:  # type: ignore
        with tempfile.NamedTemporaryFile(
            dir=self._file_settings.tmp_fixtures_dir,
            prefix=f"{context.test_run.id}_",
            suffix=self._file_settings.fixture_suffix,
            mode="w",
        ) as file:
            joined_content = "\n".join(self._project_settings.fixture_content)
            data = joined_content.format(feature_file_path=feature_file.name)

            logger.debug("Fixture file:\n%s", data)
            self._write_data(file=file, data=data, entity_name="fixture")
            yield file

    def produce_feature_file(self, context: TestExecutorContext) -> Path:
        feature_file_path = (
            self._feature_extractor.feature_type_to_dir_mapping[context.feature.feature_type.name]
            / context.feature.file_path
        )
        logger.info("Write scenario to feature file '%s'...", feature_file_path.as_posix())

        scenario_text = self._scenario_compiler.compile(context=context)
        logger.debug(scenario_text)
        feature_file_path.parent.mkdir(parents=True, exist_ok=True)
        feature_file_path.write_text(scenario_text)

        logger.info("Scenario has successfully written")
        return feature_file_path
